"""Functions to read configuration.
"""
import configparser
import logging
import os
import re
import sys
from collections import defaultdict
from glob import glob
from typing import List, Optional

SECTION = "odoo_scripts"
OTHER_SOURCES_KEY = "other_sources"
ODOO_7 = "odoo7"
"""Value for odoo_type representing Odoo 7.0"""
ODOO_8 = "odoo8"
"""Value for odoo_type representing Odoo 8.0"""
ODOO_9 = "odoo9"
""".. warning: This version is not handled"""
ODOO_10 = "odoo10"
"""Value for odoo_type representing Odoo 10.0"""
ODOO_11 = "odoo11"
"""Value for odoo_type representing Odoo 11.0"""
ODOO_12 = "odoo12"
""".. warning: This version is not handled"""
ODOO_13 = "odoo13"
"""Value for odoo_type representing Odoo 13.0"""
ODOO_14 = "odoo14"
""".. warning: This version is not handled"""
ODOO_15 = "odoo15"
"""Value for odoo_type representing Odoo 15.0"""

_logger = logging.getLogger(__name__)


class Configuration:
    """
    At the moment, only read from `setup.cfg`.
    """

    def __init__(
        self,
        path: str = None,
        blacklist: str = None,
        read: str = None,
        parent_configuration=None,
        registry: str = None,
    ):
        """
        :param registry: use the provided registry instead of the one in the
        configuration file
        """
        self._expanded_configuration = {}
        self.path = path
        setup_path = "setup.cfg"
        if path:
            setup_path = os.path.join(path, setup_path)
        config_parser = configparser.ConfigParser()
        if not os.path.exists(setup_path):
            _logger.warning("Missing %s", setup_path)
        else:
            config_parser.read(setup_path)
        section = config_parser[SECTION] if SECTION in config_parser else {}

        # read expanded configurations information
        expanded_info = defaultdict(dict)
        for key in section:
            if key.startswith("expand."):
                remaining_key = key.split(".")[1:]
                expanded_info[remaining_key[0]][remaining_key[1]] = section.get(key)

        for name, value in expanded_info.items():
            value["parent_configuration"] = self
            if "path" not in value:
                _logger.error("Expansion without path for %s", name)
                raise Exception("Expansion without path for %s" % name)
            # put in relative_path the relative path between this and the
            # parent.
            # this allows blacklisting odoo-mint/theme when expanding
            # mint hr.
            if parent_configuration:
                # difference between this path and the parents
                relative_path = os.path.join(
                    os.path.relpath(
                        path,
                        parent_configuration.path if parent_configuration.path else "",
                    ),
                    value["path"],
                )
            else:
                relative_path = value["path"]
            if blacklist and re.match(blacklist, relative_path):
                _logger.info(
                    "Ignoring directory '%s' due to blacklisting",
                    os.path.join(path, value["path"]) if path else value["path"],
                )
            else:
                if path:
                    value["path"] = os.path.join(path, value["path"])
                if blacklist:
                    if "blacklist" in value:
                        # combine the two regexp
                        value["blacklist"] = "{}|{}".format(
                            value["blacklist"], blacklist
                        )
                    else:
                        value["blacklist"] = blacklist
                # read expanded configurations
                self._expanded_configuration[name] = Configuration(**value)

        def set_from_glob_values(
            base: Optional[str], glob_values: List[str], only_directories: bool
        ) -> set:
            result_set = set()
            for glob_value in glob_values:
                if base:
                    glob_value = os.path.join(base, glob_value)
                for file in glob(glob_value):
                    if not only_directories or os.path.isdir(file):
                        result_set.add(file)
            return result_set

        def toread(key: str) -> bool:
            """Return true if value of key is to be read from the file"""
            if read:
                return key in read.split(",")
            return True

        # those keys are path to files (glob compatible)
        for key in ("modules", "dependencies", OTHER_SOURCES_KEY):
            set_values = set()
            if toread(key):
                # add any expanded values
                for config in self._expanded_configuration.values():
                    set_values.update(getattr(config, key))
                # get the values from this configuration
                values = section.get(key, "").split()
                # XXX eventually add option to not include path
                set_values.update(
                    set_from_glob_values(path, values, key != OTHER_SOURCES_KEY)
                )
            setattr(self, key, list(set_values))

        self.addons_path = {os.path.dirname(module) for module in self.modules}
        """Set of directory containing modules"""

        def key_format(a_key: str) -> str:
            """Replace invalid characters in an attribute"""
            return a_key.replace("-", "_").replace(".", "_")

        # those keys are list of values
        for key in ("module_list", "module_list_tests", "pg.extensions"):
            value = []
            formatted_key = key_format(key)
            if toread(key):
                # add any expanded values
                for config in self._expanded_configuration.values():
                    value.extend(getattr(config, formatted_key))
                # add local values
                value.extend(section.get(key, "").split())
            # then set it
            setattr(self, formatted_key, value)

        # those keys are single string values, no expand

        for key in (
            "db_user",
            "db_password",
            "load-language",
            "duplicate_repo",
            "marabunta_migration_file",
        ):
            setattr(self, key_format(key), section.get(key, None))

        def read_expanded(key: str, default_value=None):
            """Read from local version, or from expanded if value is found.
            Raise an exception if more than one value is found."""
            value = default_value
            if toread(key):
                found = False
                for config in self._expanded_configuration.values():
                    if hasattr(config, key):
                        if found:
                            raise Exception(
                                "Key %s to read in several expanded "
                                "configurations" % key
                            )
                        value = getattr(config, key)
                        found = True
                if key in section:
                    local_value = section.get(key, default_value)
                    if found and value != default_value and local_value != value:
                        raise Exception(
                            "Key %s in expanded configuration and locally set"
                            " with different values" % key
                        )
                    value = local_value
            if toread(key):
                setattr(self, key, value)

        self.registry = (
            section.get("registry", "registry.xcg.io") if registry is None else registry
        )
        project_path = os.path.realpath(".")
        self.image = section.get("image", os.path.basename(project_path))
        self.repository = f"{self.registry}/{self.image}"
        """Registry and image name combination, read from the configuration file"""
        self.local_tag = "dev"
        """Tag used for local operations"""
        # Does not use repository as it is not meant to be pushed
        self.local_image = self.image
        self.local_image_name = f"{self.local_image}:{self.local_tag}"
        """Full image name used for local operations (locally built, used for test and
        running)"""
        read_expanded("odoo_type", "odoo7")
        if not hasattr(self, "odoo_type"):
            # if path is set, odoo_type might not be in the values to read
            # so no message
            if not path:
                _logger.warning("Missing odoo_type")
        elif self.odoo_type not in (
            "odoo7",
            "odoo8",
            "odoo10",
            "odoo11",
            "odoo13",
            ODOO_15,
        ):
            if path:
                _logger.info(
                    "Unexpected odoo_type: %s in %s",
                    self.odoo_type,
                    self.path,
                )
            else:
                _logger.warning("Unexpected odoo_type: %s", self.odoo_type)
        read_expanded("postgresql_version", "9.6")
        self.start_py3o = section.get("start_py3o", "no") in ("yes", "true")
        self.start_rednerd = section.get("start_rednerd", "no") in (
            "yes",
            "true",
        )
        read_expanded("rednerd_version", "latest")
        read_expanded("rednerd_repository", None)


class Singleton(type):
    """Singleton metaclass. From
    https://stackoverflow.com/questions/6760685/creating-a-singleton-in-python
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Config(Configuration, metaclass=Singleton):
    """Singleton configuration."""

    def __init__(self):
        super().__init__()


def main(argv=None):
    """Display configuration. Useful for development of odoo_scripts."""
    c = Config()
    for key, value in c.__dict__.items():
        print(key, "=", value)
    return 0


if __name__ == "__main__":
    return_code = main(sys.argv[1:])
    if return_code:
        sys.exit(return_code)
