#!/usr/bin/env python3
# vim: set shiftwidth=4 softtabstop=4:
"""Run flake8 in a docker
"""
import argparse
import logging
import os
import sys

from .config import ODOO_15, Config
from .docker_client import DockerClient
from .parsing import apply, basic_parser

__version__ = "1.2.0"
__date__ = "2020-09-10"
__updated__ = "2022-01-27"

_logger = logging.getLogger(__name__)


def __parser():
    program_version = __version__
    program_build_date = str(__updated__)
    program_version_message = "%%(prog)s %s (%s)" % (
        program_version,
        program_build_date,
    )
    program_shortdesc = __doc__.split(".", maxsplit=1)[0]
    program_license = """%s

      Created by Vincent Hatakeyama on %s.
      Copyright 2020, 2022 XCG Consulting. All rights reserved.

      Licensed under the MIT License

      Distributed on an "AS IS" basis without warranties
      or conditions of any kind, either express or implied.

    USAGE
    """ % (
        program_shortdesc,
        str(__date__),
    )
    parser = basic_parser(program_license, program_version_message)
    return parser


def parser_add_flake8_group(parser: argparse.ArgumentParser):
    """Add flake8 option (--flake8/--no-flake8) to the given parser.
    Defaults to not running flake8."""
    flake8_group = parser.add_mutually_exclusive_group()
    flake8_group.add_argument(
        "--flake8",
        help="Run flake8",
        action="store_true",
        dest="flake8",
        default=False,
    )
    flake8_group.add_argument(
        "--no-flake8",
        help="Do not run flake8 [default]",
        action="store_false",
        dest="flake8",
        default=False,
    )


def apply_flake8(namespace: argparse.Namespace, odoo_type: str):
    """Run flake8 if the option was set."""
    if namespace.flake8:
        flake8(odoo_type)


def main(argv=None):
    """Copy modules for a build, callable version that parses arguments"""
    parser = __parser()
    nmspc = parser.parse_args(argv)
    apply(nmspc)
    config = Config()
    odoo_type = config.odoo_type
    return flake8(odoo_type)


def flake8(odoo_type: str, pull: bool = True):
    """Run flake8"""
    # determine image to use based on odoo version
    repository = "quay.orus.io/odoo/odoo"
    if odoo_type in ("odoo7", "odoo8", "odoo9", "odoo10"):
        repository = "xcgd/flake8"
        tag = "2"
    elif odoo_type == ODOO_15:
        tag = "15.0"
    else:
        repository = "xcgd/flake8"
        tag = "3"

    _logger.info("Running flake8")
    pwd = os.environ["PWD"]
    return DockerClient.run(
        repository,
        tag,
        {
            "entrypoint": "flake8",
            "command": ["/mnt"],
            "volumes": {pwd: {"bind": "/mnt", "mode": "ro"}},
            "working_dir": "/mnt",
        },
        pull,
        {"/mnt": pwd},
    )


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))
