#!/usr/bin/env python3
# vim: set shiftwidth=4 softtabstop=4:
"""Run isort in a docker
"""
import argparse
import logging
import os
import sys

from .config import ODOO_13, ODOO_15, Config
from .docker_client import DockerClient
from .parsing import apply, basic_parser

__version__ = "2.1.0"
__date__ = "2020-09-10"
__updated__ = "2022-01-27"

_logger = logging.getLogger(__name__)
_ISORT_DEST = "isort"
"""Name of destination variable used for isort in parsing"""

CHECK_MODE = "check"
"""Run isort to check"""
DIFF_MODE = "diff"
"""Run isort to show differences"""


def __parser():
    program_version = __version__
    program_build_date = str(__updated__)
    program_version_message = "%%(prog)s %s (%s)" % (
        program_version,
        program_build_date,
    )
    program_shortdesc = __doc__.split(".", maxsplit=1)[0]
    program_license = """%s

      Created by Vincent Hatakeyama on %s.
      Copyright 2020, 2022 XCG Consulting. All rights reserved.

      Licensed under the MIT License

      Distributed on an "AS IS" basis without warranties
      or conditions of any kind, either express or implied.

    USAGE
    """ % (
        program_shortdesc,
        str(__date__),
    )
    parser = basic_parser(program_license, program_version_message)
    return parser


def parser_add_isort_group(parser: argparse.ArgumentParser):
    """Add isort option (--isort/--no-isort) to the given parser.
    Defaults to not using isort."""
    isort_group = parser.add_mutually_exclusive_group()
    isort_group.add_argument(
        "--isort-check",
        help="Run isort (check only)",
        action="store_const",
        dest=_ISORT_DEST,
        const=CHECK_MODE,
    )
    isort_group.add_argument(
        "--isort-diff",
        help="Run isort (check only)",
        action="store_const",
        dest=_ISORT_DEST,
        const=DIFF_MODE,
    )
    isort_group.add_argument(
        "--no-isort",
        help="Do not run isort [default]",
        action="store_const",
        dest=_ISORT_DEST,
        const=None,
    )


def apply_isort(namespace: argparse.Namespace, odoo_type: str):
    """Run isort if the option was set."""
    if _ISORT_DEST in namespace and getattr(namespace, _ISORT_DEST):
        return isort(getattr(namespace, _ISORT_DEST), odoo_type)
    return 0


def main(argv=None):
    """Copy modules for a build, callable version that parses arguments"""
    parser = __parser()
    nmspc = parser.parse_args(argv)
    apply(nmspc)
    config = Config()
    odoo_type = config.odoo_type
    return isort(DIFF_MODE, odoo_type)


def isort(mode: str, odoo_type: str, pull: bool = True):
    """Run isort"""
    if mode == CHECK_MODE:
        command = ["--check"]
    if mode == DIFF_MODE:
        command = ["--check", "--diff"]

    # determine image to use based on odoo version
    repository = "quay.orus.io/odoo/odoo"
    if odoo_type == ODOO_15:
        tag = "15.0"
        # append the directory, needed with isort≥5.0.0
        command.append(".")
    elif odoo_type == ODOO_13:
        tag = "13.0"
    else:
        repository = "xcgd/isort"
        tag = "odoo"

    _logger.info("Running isort")
    pwd = os.environ["PWD"]
    return DockerClient.run(
        repository,
        tag,
        {
            "entrypoint": "isort",
            "user": "root",
            "command": command,
            "volumes": {
                pwd: {
                    "bind": "/mnt",
                    "mode": "ro" if mode in (CHECK_MODE, DIFF_MODE) else "rw",
                }
            },
            "working_dir": "/mnt",
        },
        pull,
        {"/mnt": pwd},
    )


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))
