#!/usr/bin/env python3
"""Run a rednerd docker with volume based on project name.

Documentation: http://127.0.0.1:12345/docs if the server is started with port 12345
"""
import argparse
import atexit
import logging
import os
import sys
import tempfile
import time
from multiprocessing import Process
from typing import Optional
from urllib.parse import quote

import docker
import requests
import requests_unixsocket
from docker.types import Mount

from .config import Config
from .docker_client import DockerClient
from .parsing import apply, basic_parser

_logger = logging.getLogger(__name__)

__version__ = "1.1.1"
__date__ = "2021-01-14"
__updated__ = "2022-09-12"

MIGRATE = "migrate"
ADMIN_PASSWORD = "admin-password"
SERVE = "serve"


class DockerRednerd:
    """
    This class allow to start the rednerd server
    """

    def __init__(
        self,
        docker_client: docker.client.DockerClient,
        project_name: str,
        repository: str,
        db_user: str,
        db_password: Optional[str],
        database: str,
        db_host: str,
        db_port: Optional[str] = None,
        attach: bool = True,
        stop_at_exit: bool = True,
        redner_user: Optional[str] = None,
        redner_password: Optional[str] = None,
        redner_admin_password: Optional[str] = None,
        redner_port: Optional[int] = None,
        use_socket: bool = False,
        version: str = "latest",
        force_pull: bool = False,
        log_level: str = "info",
    ):
        """
        :param docker_client: instance of docker client
        :param project_name: name of the project, used for some volume and container
        :param repository: repository to use for the redner image
        :param db_user: redner database user
        :param db_password: redner database password
        :param database: database name for rednerd, the database must already be created
        :param db_host: redner database host (unix socket or hostname)
        :param db_port: database port
        :param redner_port: Port to use for http, if None, no port is given to rednerd
        :param use_socket: Set to true to make rednerd use a Unix socket
        :param force_pull: Indicate to pull the latest version of the specified image,
            otherwise the image will only be pulled if not present
        """
        assert all(
            argument is not None
            for argument in (
                docker_client,
                project_name,
                repository,
                db_user,
                database,
                db_host,
            )
        ), "Missing required argument"
        # db_password can be none, in that case expect that any password would work
        super().__init__()
        self.name = f"rednerd_{project_name}"
        self.repository = repository
        self.version = version
        self.attach = attach
        self.base_image = f"{self.repository}:{self.version}"
        self.stop_at_exit = stop_at_exit
        self.docker_cli = docker_client
        self.redner_user = redner_user
        self.redner_password = redner_password
        self.redner_admin_password = redner_admin_password
        self.db_is_socket_path = db_host.startswith("/")
        self.db_host = db_host
        dsn = f"host={db_host} dbname={database} user={db_user} password={db_password}"
        if db_port:
            dsn += f" port={db_port}"
        self.environment = {
            "REDNERD_ENVIRONMENT": "local",
            "REDNERD_LOG_FORMAT": "pretty",
            "REDNERD_DSN": dsn,
            "REDNERD_LOG_LEVEL": log_level,
            # errors when not defined when using standard docker
            "NO_CONFD": "true",
        }
        if redner_port is not None:
            self.redner_port = str(redner_port)
        # create a socket directory
        if use_socket:
            self.socket_dir = tempfile.TemporaryDirectory(
                prefix="odoo_scripts_rednerd_socket-"
            ).name
            """Rednerd socket’s directory."""
            self.socket_path = os.path.join(self.socket_dir, "rednerd.sock")

        # Retrieve the container if it already exists
        try:
            self.container = self.docker_cli.containers.get(self.name)
        except docker.errors.NotFound:
            self.container = None
        self.force_pull = force_pull
        self.api_key = None
        """Store the API key"""
        self._log_process = None

    def start(self):
        """Start rednerd container"""
        if self.force_pull or not self.docker_cli.images.list(self.base_image):
            _logger.info("Pulling %s", self.base_image)
            self.pull_image()
        if self.container is not None:
            # probably to make it work as a restart
            _logger.info("Container %s already exist", self.name)
            # stop and remove
            self.stop_and_remove()
        # base parameters for container creation
        create_param = dict(
            name=self.name,
            ports=dict(),
            environment=self.environment,
            volumes=dict(),
            mounts=[],
            network_mode="host",
        )
        if self.db_is_socket_path:
            create_param["mounts"].append(Mount(self.db_host, self.db_host, "bind"))
        # migrate
        _logger.info("migrate")
        DockerClient.run(
            repository=self.repository,
            tag=self.version,
            run_kwargs=dict(create_param, command=[MIGRATE]),
        )
        # setup admin password if provided
        if self.redner_admin_password is not None:
            _logger.info("admin-password")
            DockerClient.run(
                repository=self.repository,
                tag=self.version,
                run_kwargs=dict(
                    create_param, command=[ADMIN_PASSWORD, self.redner_admin_password]
                ),
            )
        # serve
        create_param["image"] = self.base_image
        create_param["command"] = [SERVE, "--token-secret", self.name]
        schemes = []
        if hasattr(self, "redner_port"):
            create_param["command"].extend(["--port", str(self.redner_port)])
            schemes.append("http")
        if hasattr(self, "socket_path"):
            create_param["command"].extend(["--socket-path", self.socket_path])
            schemes.append("unix")
            _logger.info("Using socket %s", self.socket_path)
        if schemes:
            create_param["command"].extend(["--scheme", ",".join(schemes)])
        if hasattr(self, "redner_port"):
            # useless when using --network=host
            create_param["ports"][f"{self.redner_port}/tcp"] = self.redner_port
            _logger.info("Using port %d", self.redner_port)
        if hasattr(self, "socket_dir"):
            create_param["volumes"][self.socket_dir] = {
                "bind": self.socket_dir,
                "mode": "rw",
            }
        self.container = self.docker_cli.containers.create(**create_param)
        _logger.info("Starting container %s", self.name)
        _logger.debug(create_param)
        self.container.start()
        session = None
        base_url = None
        if hasattr(self, "socket_path"):
            session = requests_unixsocket.Session()
            base_url = "http+unix://{}".format(quote(self.socket_path, safe=""))
        elif hasattr(self, "redner_port"):
            session = requests.sessions.Session()
            base_url = f"http://localhost:{self.redner_port}"
        # if there is neither port or socket, no session nor base_url
        if not self.attach:
            process = Process(target=DockerClient.print_log, args=(self.container,))
            process.start()
            self._log_process = process
        # wait for redner to be started
        if hasattr(self, "socket_path"):
            _logger.info("Waiting for redner socket")
            while not os.path.exists(self.socket_path):
                time.sleep(0.1)
        elif hasattr(self, "redner_port"):
            _logger.info("Wait for port to be available")
            ready = False
            while not ready:
                try:
                    session.get(base_url)
                    ready = True
                except requests.exceptions.ConnectionError:
                    pass
                time.sleep(0.2)
        else:
            _logger.warning("No port nor socket to wait for")
        # change socket permissions, current version has too much restrictive values
        # it might be due to requests_unixdomain that want more than available
        if hasattr(self, "socket_path"):
            _logger.info("Changing socket permission")
            DockerClient.exec(
                self.container, ["chmod", "-R", "777", self.socket_path], "root"
            )

        if (
            base_url is not None
            and session is not None
            and self.redner_admin_password is not None
        ):
            _logger.info("Log in as admin")
            # auth
            request = session.post(
                f"{base_url}/api/v1/auth/authenticate",
                json={
                    "username": "admin",
                    "password": self.redner_admin_password,
                    "set-cookie": True,
                },
            )
            request.raise_for_status()
            cookie_jar = request.cookies
            # check if user exists
            request = session.get(
                f"{base_url}/api/v1/user/",
                # json={"username": self.redner_user},
                cookies=cookie_jar,
            )
            request.raise_for_status()
            user_missing = True
            for user in request.json():
                if user["username"] == self.redner_user:
                    user_missing = False
                    continue
            if user_missing:
                _logger.info("Creating user %s", self.redner_user)
                request = session.post(
                    f"{base_url}/api/v1/user",
                    json={"username": self.redner_user, "is-admin": False},
                    cookies=cookie_jar,
                )
                request.raise_for_status()
            # change user password
            _logger.info("Creating %s password", self.redner_user)
            request = session.post(
                f"{base_url}/api/v1/user/{self.redner_user}/password",
                json={"password": self.redner_password},
                cookies=cookie_jar,
            )
            request.raise_for_status()
            _logger.info("Grabbing API key for %s", self.redner_user)
            request = session.get(
                f"{base_url}/api/v1/user/{self.redner_user}/apikey",
                json={},
                cookies=cookie_jar,
            )
            request.raise_for_status()
            api_keys_list = request.json()
            if len(api_keys_list) > 0:
                self.api_key = api_keys_list[0]["value"]
            if self.api_key is None:
                _logger.info("No API key found, creating one")
                # create a new token otherwise
                request = session.post(
                    f"{base_url}/api/v1/user/{self.redner_user}/apikey",
                    json={
                        "name": "docker_rednerd",
                        "description": "Odoo scripts created API KEY",
                    },
                    cookies=cookie_jar,
                )
                self.api_key = request.json()["value"]

        if self.stop_at_exit:
            atexit.register(self.stop_and_remove)

        if self.attach:
            DockerClient.print_log(self.container)

        return True

    def stop_and_remove(self):
        """Stop and remove the rednerd container"""
        try:
            self.stop()  # call rednerd stop func
            self.remove()  # call rednerd remove func
        except docker.errors.NotFound as exception:
            _logger.info("Stop and remove rednerd container, got: %s", exception)
            raise exception

    def stop(self):
        """
        Stop a rednerd container.
        """
        try:
            if not self.is_running:
                _logger.warning("Container %s is not running", self.name)
            _logger.info("Stopping %s container", self.name)
            self.container.stop()
            self.container.wait()
        except docker.errors.NotFound as exception:
            _logger.error("Stopping Redner container, got: %s", exception)
            raise
        if self._log_process and self._log_process.is_alive():
            _logger.info("Waiting for log process to finish")
            self._log_process.join()

    def remove(self):
        """Remove the rednerd container created"""
        try:
            _logger.info("Removing %s container", self.name)
            self.container.remove()
        except docker.errors.NotFound as exception:
            _logger.error("Removing rednerd container, got: %s", exception)
            raise exception

    def pull_image(self):
        """
        Pulls rednerd image
        """
        self.docker_cli.images.pull(repository=self.repository, tag=self.version)

    @property
    def is_running(self):
        """
        Verify the rednerd container status
        :return: Boolean if the status is running
        """
        if self.container is None:
            return False
        result = self.docker_cli.containers.list(filters={"id": self.container.id})
        return result and result[0].status == "running"

    @property
    def is_exited(self):
        """
        Verify the rednerd container status
        :return: Boolean if the status is exited
        """
        if self.container is None:
            return True
        result = self.docker_cli.containers.list(filters={"id": self.container.id})
        return result and result[0].status == "exited"


def __parser() -> argparse.ArgumentParser:
    """Return a parser for docker_rednerd"""
    program_version_message = f"%(prog)s {__version__} ({__updated__})"
    program_short_description = __doc__.split(".")[0]
    program_license = f"""{program_short_description}

      Created by Oury Balde on {__date__}.
      Copyright 2021 XCG Consulting. All rights reserved.

      Licensed under the MIT License

      Distributed on an "AS IS" basis without warranties
      or conditions of any kind, either express or implied.

    USAGE
    """
    project_path = os.path.realpath(".")
    project_name = os.path.basename(project_path)

    parser = basic_parser(program_license, program_version_message)
    parser.add_argument(
        "--db-user",
        help="Database user [default: %(default)s]",
    )
    parser.add_argument(
        "--db-password",
        help="Database password [default: %(default)s]",
    )
    parser.add_argument(
        "--database",
        help="Database name [default: %(default)s]",
    )
    parser.add_argument(
        "--db-host",
        help="Unix socket or hostname to access the database [default: %(default)s]",
    )
    parser.add_argument(
        "--db-port",
        help="Database port [default: %(default)s]",
        default=None,
    )
    parser.add_argument(
        "--name",
        help="Base container name to use [default: %(default)s]",
        default=project_name,
    )
    parser.add_argument(
        "--user",
        help="Redner user to use [default: %(default)s]",
        default=None,
    )
    parser.add_argument(
        "--password",
        help="Redner password to use [default: %(default)s]",
        default=None,
    )
    parser.add_argument(
        "--admin-password",
        help="Redner admin password to use [default: %(default)s]",
        default=None,
    )
    parser.add_argument(
        "-p",
        "--port",
        help="Redner port to use, use a socket if empty [default: %(default)s]",
        default=None,
        type=int,
    )
    parser.add_argument(
        "--use-socket",
        help="Specify Redner socket to use [default: %(default)s]",
        action="store_true",
        default=False,
    )
    return parser


def main(argv=None):
    """Run redner docker using command line arguments"""
    parser = __parser()
    namespace = parser.parse_args(argv)
    apply(namespace)
    project_name = namespace.name
    db_user = namespace.db_user
    db_password = namespace.db_password
    database = namespace.database
    db_host = namespace.db_host
    db_port = namespace.db_port
    user = namespace.user
    password = namespace.password
    admin_password = namespace.admin_password
    port = namespace.port
    use_socket = namespace.use_socket

    # get rednerd version and repository from config
    config = Config()
    rednerd_version = config.rednerd_version
    rednerd_repository = config.rednerd_repository

    # Instantiate docker rednerd
    docker_rednerd = DockerRednerd(
        DockerClient.client,
        project_name,
        rednerd_repository,
        db_user=db_user,
        db_password=db_password,
        database=database,
        db_host=db_host,
        db_port=db_port,
        attach=True,
        redner_user=user,
        redner_password=password,
        redner_admin_password=admin_password,
        redner_port=port,
        version=rednerd_version,
        use_socket=use_socket,
    )
    docker_rednerd.start()

    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))
